from abc import ABC, abstractmethod
from enum import Enum
from typing import List

from shor.quantum import QuantumCircuit
from shor.utils.qbits import int_from_bit_string, int_to_bit_string


class Result(ABC):
    @property
    @abstractmethod
    def counts(self):
        pass

    @property
    @abstractmethod
    def sig_bits(self):
        pass

    def get(self, key, default=0):
        if isinstance(key, str):
            idx = int_from_bit_string(key)
        else:
            idx = key

        return self.counts.get(idx, default)

    def __getitem__(self, item):
        return self.get(item)

    def __repr__(self):
        return repr({int_to_bit_string(k, sig_bits=self.sig_bits): v for k, v in self.counts.items()})


class JobStatusCode(Enum):
    COMPLETED = ("completed",)
    ERROR = ("error",)
    RUNNING = ("running",)
    WAITING = "waiting"


class JobStatus(object):
    def __init__(self, code: JobStatusCode, message: str = "", api_error_code: int = None):
        self.code = code
        self.message = message
        self.api_error_code = api_error_code


class Job(ABC):
    @property
    @abstractmethod
    def result(self):
        pass

    @property
    @abstractmethod
    def status(self):
        pass


class Provider(ABC):
    @property
    @abstractmethod
    def jobs(self) -> List[Job]:
        pass

    @abstractmethod
    def login(self, token: str) -> bool:
        pass

    @abstractmethod
    def logout(self):
        pass

    @abstractmethod
    def run(self, circuit: QuantumCircuit, times: int) -> Job:
        pass



from qiskit.providers.ibmq import IBMQ

from shor.providers.qiskit.base import QiskitProvider

DEFAULT_BACKEND = "qasm_simulator"


class IBMQProvider(QiskitProvider):
    def __init__(self, **config):
        if not IBMQ.active_account():
            IBMQ.load_account()

        provider = IBMQ.providers()[0] if IBMQ.providers() else None
        config["provider"] = provider

        if "backend" in config:
            if not isinstance(config["backend"], str):
                config["backend"] = provider.get_backend(DEFAULT_BACKEND) if provider and provider.backends() else None

        super().__init__(**config)



from typing import List

from qiskit import Aer, execute

from shor.providers.base import Job, Provider, Result
from shor.quantum import QC
from shor.transpilers.qiskit import to_qiskit_circuit
from shor.utils.qbits import int_from_bit_string

DEFAULT_BACKEND = Aer.get_backend("qasm_simulator")
DEFAULT_PROVIDER = Aer


class QiskitResult(Result):
    def __init__(self, qiskit_result):
        self.qiskit_result = qiskit_result

    @property
    def counts(self):
        return {int_from_bit_string(k.split(" ")[0]): v for k, v in self.qiskit_result.get_counts().items()}

    @property
    def sig_bits(self):
        measurement_bases = list(self.qiskit_result.get_counts().keys())
        return len(measurement_bases[0]) if measurement_bases else 0


class QiskitJob(Job):
    def __init__(self, qiskit_job):
        self.qiskit_job = qiskit_job

    @property
    def status(self):
        return self.qiskit_job.status()

    @property
    def result(self) -> QiskitResult:
        return QiskitResult(self.qiskit_job.result())


class QiskitProvider(Provider):
    def __init__(self, **config):
        self.provider_delegate = config.get("provider", DEFAULT_PROVIDER)

        if "backend" in config:
            self.load_backend(config["backend"])
        else:
            self.backend = config.get("backend", DEFAULT_BACKEND)

    def backends(self):
        return self.provider_delegate.backends()

    def load_backend(self, backend: str):
        self.backend = self.provider_delegate.get_backend(backend)

    @property
    def jobs(self) -> List[Job]:
        return list(map(lambda j: QiskitJob(j), self.backend.get_jobs()))

    def login(self, token: str, remember: bool = False, **kwargs) -> None:
        self.backend.enable_account(token, **kwargs)
        if remember:
            self.backend.save_account(token, **kwargs)

    def logout(self) -> None:
        self.backend.disable_account()
        self.backend.delete_account()

    def run(self, circuit: QC, times: int) -> QiskitJob:
        job = execute(to_qiskit_circuit(circuit), self.backend, shots=times)

        return QiskitJob(job)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['shor',
 'shor.algorithms',
 'shor.providers',
 'shor.providers.qiskit',
 'shor.transpilers',
 'shor.utils']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.3.3,<4.0.0', 'numpy>=1.19.4,<2.0.0', 'qiskit>=0.23.1,<0.24.0']

setup_kwargs = {
    'name': 'shor',
    'version': '0.0.4',
    'description': 'Quantum Computing for Humans',
    'long_description': None,
    'author': 'Collin Overbay - shor.dev',
    'author_email': 'shordotdev@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
